# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""An intent parsing service using the Adapt parser."""
from functools import lru_cache
from threading import Lock
from typing import List, Tuple, Optional

from ovos_adapt.engine import IntentDeterminationEngine
from ovos_bus_client.message import Message
from ovos_bus_client.session import IntentContextManager as ContextManager, \
    SessionManager
from ovos_config.config import Configuration
from ovos_plugin_manager.templates.pipeline import IntentMatch, PipelinePlugin
from ovos_utils import flatten_list
from ovos_utils.log import LOG


def _entity_skill_id(skill_id):
    """Helper converting a skill id to the format used in entities.

    Arguments:
        skill_id (str): skill identifier

    Returns:
        (str) skill id on the format used by skill entities
    """
    skill_id = skill_id[:-1]
    skill_id = skill_id.replace('.', '_')
    skill_id = skill_id.replace('-', '_')
    return skill_id


class AdaptPipeline(PipelinePlugin):
    """Intent service wrapping the Adapt intent Parser."""

    def __init__(self, config=None):
        core_config = Configuration()
        self.config = config or core_config.get("context", {})  # legacy mycroft-core path
        self.lang = core_config.get("lang", "en-us")
        langs = core_config.get('secondary_langs') or []
        if self.lang not in langs:
            langs.append(self.lang)

        self.engines = {lang: IntentDeterminationEngine()
                        for lang in langs}

        self.lock = Lock()
        self.max_words = 50  # if an utterance contains more words than this, don't attempt to match

        # TODO sanitize config option
        self.conf_high = self.config.get("conf_high") or 0.65
        self.conf_med = self.config.get("conf_med") or 0.45
        self.conf_low = self.config.get("conf_low") or 0.25

    @property
    def context_keywords(self):
        LOG.warning(
            "self.context_keywords has been deprecated and is unused, use self.config.get('keywords', []) instead")
        return self.config.get('keywords', [])

    @context_keywords.setter
    def context_keywords(self, val):
        LOG.warning(
            "self.context_keywords has been deprecated and is unused, edit mycroft.conf instead, setter will be ignored")

    @property
    def context_max_frames(self):
        LOG.warning(
            "self.context_keywords has been deprecated and is unused, use self.config.get('max_frames', 3) instead")
        return self.config.get('max_frames', 3)

    @context_max_frames.setter
    def context_max_frames(self, val):
        LOG.warning(
            "self.context_max_frames has been deprecated and is unused, edit mycroft.conf instead, setter will be ignored")

    @property
    def context_timeout(self):
        LOG.warning("self.context_timeout has been deprecated and is unused, use self.config.get('timeout', 2) instead")
        return self.config.get('timeout', 2)

    @context_timeout.setter
    def context_timeout(self, val):
        LOG.warning(
            "self.context_timeout has been deprecated and is unused, edit mycroft.conf instead, setter will be ignored")

    @property
    def context_greedy(self):
        LOG.warning(
            "self.context_greedy has been deprecated and is unused, use self.config.get('greedy', False) instead")
        return self.config.get('greedy', False)

    @context_greedy.setter
    def context_greedy(self, val):
        LOG.warning(
            "self.context_greedy has been deprecated and is unused, edit mycroft.conf instead, setter will be ignored")

    @property
    def context_manager(self):
        LOG.warning("context_manager has been deprecated, use Session.context instead")
        sess = SessionManager.get()
        return sess.context

    @context_manager.setter
    def context_manager(self, val):
        LOG.warning("context_manager has been deprecated, use Session.context instead")
        assert isinstance(val, ContextManager)
        sess = SessionManager.get()
        sess.context = val

    def update_context(self, intent):
        """Updates context with keyword from the intent.

        NOTE: This method currently won't handle one_of intent keywords
              since it's not using quite the same format as other intent
              keywords. This is under investigation in adapt, PR pending.

        Args:
            intent: Intent to scan for keywords
        """
        LOG.warning("update_context has been deprecated, use Session.context.update_context instead")
        sess = SessionManager.get()
        ents = [tag['entities'][0] for tag in intent['__tags__'] if 'entities' in tag]
        sess.context.update_context(ents)

    def match_high(self, utterances: List[str],
                   lang: Optional[str] = None,
                   message: Optional[Message] = None):
        """Intent matcher for high confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        match = self.match_intent(tuple(utterances), lang, message.serialize())
        if match and match.intent_data.get("confidence", 0.0) >= self.conf_high:
            return match
        return None

    def match_medium(self, utterances: List[str],
                     lang: Optional[str] = None,
                     message: Optional[Message] = None):
        """Intent matcher for medium confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        match = self.match_intent(tuple(utterances), lang, message.serialize())
        if match and match.intent_data.get("confidence", 0.0) >= self.conf_med:
            return match
        return None

    def match_low(self, utterances: List[str],
                  lang: Optional[str] = None,
                  message: Optional[Message] = None):
        """Intent matcher for low confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        match = self.match_intent(tuple(utterances), lang, message.serialize())
        if match and match.intent_data.get("confidence", 0.0) >= self.conf_low:
            return match
        return None

    @lru_cache(maxsize=3)  # NOTE - message is a string because of this
    def match_intent(self, utterances: Tuple[str],
                     lang: Optional[str] = None,
                     message: Optional[str] = None):
        """Run the Adapt engine to search for an matching intent.

        Args:
            utterances (iterable): utterances for consideration in intent 
                    matching. As a practical matter, a single utterance will 
                    be passed in most cases. But there are instances, such as
                    streaming STT that could pass multiple. Each utterance is 
                    represented as a tuple containing the raw, normalized, and
                    possibly other variations of the utterance.
            limit (float): confidence threshold for intent matching
            lang (str): language to use for intent matching
            message (Message): message to use for context

        Returns:
            Intent structure, or None if no match was found.
        """

        if message:
            message = Message.deserialize(message)
        sess = SessionManager.get(message)

        # we call flatten in case someone is sending the old style list of tuples
        utterances = flatten_list(utterances)

        utterances = [u for u in utterances if len(u.split()) < self.max_words]
        if not utterances:
            LOG.error(f"utterance exceeds max size of {self.max_words} words, skipping adapt match")
            return None

        lang = lang or self.lang
        if lang not in self.engines:
            return None

        best_intent = {}

        def take_best(intent, utt):
            nonlocal best_intent
            best = best_intent.get('confidence', 0.0) if best_intent else 0.0
            conf = intent.get('confidence', 0.0)
            skill = intent['intent_type'].split(":")[0]
            if best < conf and intent["intent_type"] not in sess.blacklisted_intents \
                    and skill not in sess.blacklisted_skills:
                best_intent = intent
                # TODO - Shouldn't Adapt do this?
                best_intent['utterance'] = utt

        for utt in utterances:
            try:
                intents = [i for i in self.engines[lang].determine_intent(
                    utt, 100,
                    include_tags=True,
                    context_manager=sess.context)]
                if intents:
                    utt_best = max(
                        intents, key=lambda x: x.get('confidence', 0.0)
                    )
                    take_best(utt_best, utt)

            except Exception as err:
                LOG.exception(err)

        if best_intent:
            ents = [tag['entities'][0] for tag in best_intent['__tags__'] if 'entities' in tag]

            sess.context.update_context(ents)

            skill_id = best_intent['intent_type'].split(":")[0]
            ret = IntentMatch(
                'Adapt', best_intent['intent_type'], best_intent, skill_id,
                best_intent['utterance']
            )
        else:
            ret = None
        return ret

    def register_vocab(self, start_concept, end_concept,
                       alias_of, regex_str, lang):
        """Register Vocabulary. DEPRECATED

        This method should not be used, it has been replaced by
        register_vocabulary().
        """
        self.register_vocabulary(start_concept, end_concept, alias_of,
                                 regex_str, lang)

    def register_vocabulary(self, entity_value, entity_type,
                            alias_of, regex_str, lang):
        """Register skill vocabulary as adapt entity.

        This will handle both regex registration and registration of normal
        keywords. if the "regex_str" argument is set all other arguments will
        be ignored.

        Argument:
            entity_value: the natural langauge word
            entity_type: the type/tag of an entity instance
            alias_of: entity this is an alternative for
        """
        if lang in self.engines:
            with self.lock:
                if regex_str:
                    self.engines[lang].register_regex_entity(regex_str)
                else:
                    self.engines[lang].register_entity(
                        entity_value, entity_type, alias_of=alias_of)

    def register_intent(self, intent):
        """Register new intent with adapt engine.

        Args:
            intent (IntentParser): IntentParser to register
        """
        for lang in self.engines:
            with self.lock:
                self.engines[lang].register_intent_parser(intent)

    def detach_skill(self, skill_id):
        """Remove all intents for skill.

        Args:
            skill_id (str): skill to process
        """
        with self.lock:
            for lang in self.engines:
                skill_parsers = [
                    p.name for p in self.engines[lang].intent_parsers if
                    p.name.startswith(skill_id)
                ]
                self.engines[lang].drop_intent_parser(skill_parsers)
            self._detach_skill_keywords(skill_id)
            self._detach_skill_regexes(skill_id)

    def _detach_skill_keywords(self, skill_id):
        """Detach all keywords registered with a particular skill.

        Arguments:
            skill_id (str): skill identifier
        """
        skill_id = _entity_skill_id(skill_id)

        def match_skill_entities(data):
            return data and data[1].startswith(skill_id)

        for lang in self.engines:
            self.engines[lang].drop_entity(match_func=match_skill_entities)

    def _detach_skill_regexes(self, skill_id):
        """Detach all regexes registered with a particular skill.

        Arguments:
            skill_id (str): skill identifier
        """
        skill_id = _entity_skill_id(skill_id)

        def match_skill_regexes(regexp):
            return any([r.startswith(skill_id)
                        for r in regexp.groupindex.keys()])

        for lang in self.engines:
            self.engines[lang].drop_regex_entity(match_func=match_skill_regexes)

    def detach_intent(self, intent_name):
        """Detatch a single intent

        Args:
            intent_name (str): Identifier for intent to remove.
        """
        for lang in self.engines:
            new_parsers = [
                p for p in self.engines[lang].intent_parsers if p.name != intent_name
            ]
            self.engines[lang].intent_parsers = new_parsers

    def shutdown(self):
        for lang in self.engines:
            parsers = self.engines[lang].intent_parsers
            self.engines[lang].drop_intent_parser(parsers)



import pickle

from ovos_adapt.engine import DomainIntentDeterminationEngine, \
    IntentDeterminationEngine

"""
Pickling is not inherently secure, and the documentation for pickle
recommends never unpickling data from an untrusted source. This makes
using it for debug reports a little dicey, but fortunately there are 
some things we can leverage to make things safer.

First, we expect the deserialized object to be an instance of 
IntentDeterminationEngine or DomainIntentDeterminationEngine as a basic
sanity check. We also leverage a custom `pickle.Unpickler` implementation
that only allows specific imports from the ovos_adapt namespace, and displays
a helpful error message if this assumption is validated.

This is a bit of security theater; folks investigating issues have to know to 
specifically use this library to hydrate any submissions, otherwise it's all 
for naught. 
"""


EXPECTED_ENGINES = set([
    IntentDeterminationEngine,
    DomainIntentDeterminationEngine,
])

SAFE_CLASSES = [
    ("ovos_adapt.engine", "IntentDeterminationEngine"),
    ("ovos_adapt.engine", "DomainIntentDeterminationEngine"),
    ("ovos_adapt.tools.text.tokenizer", "EnglishTokenizer"),
    ("ovos_adapt.tools.text.trie", "Trie"),
    ("ovos_adapt.tools.text.trie", "TrieNode"),
    ("ovos_adapt.intent", "Intent")
]


class RestrictedUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        if (module, name) not in SAFE_CLASSES:
            raise pickle.UnpicklingError("Attempted illegal import: "
                                         "{}.{}".format(module, name))
        return pickle.Unpickler.find_class(self, module, name)


def load(filename):
    """
    Load a file that contains a serialized intent determination engine.
    :param filename (str): source path
    :return: An instance of IntentDeterminationEngine or
        DomainIntentDeterminationEngine
    """
    with open(filename, 'rb') as f:
        engine = RestrictedUnpickler(f).load()
        if engine.__class__ not in EXPECTED_ENGINES:
            raise ValueError("Was expecting to instantiate an "
                             "IntentDeterminationEngine, but instead found "
                             "{}".format(engine.__class__))
        return engine


def dump(engine, filename):
    """
    Serialize an ovos_adapt Intent engine and write it to the target file.
    :param engine (IntentDeterminationEngine or DomainIntentDeterminationEngine):
    :param filename (str): destination path
    """
    with open(filename, 'wb') as f:
        pickle.dump(engine, f)



# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
from ovos_adapt.expander import BronKerboschExpander
from ovos_adapt.tools.text.trie import Trie

__author__ = 'seanfitz'


class Parser(object):
    """
    Coordinate a tagger and expander to yield valid parse results.
    """
    def __init__(self, tokenizer, tagger):
        self._tokenizer = tokenizer
        self._tagger = tagger

    def parse(self, utterance, context=None, N=1):
        """Used to find tags within utterance with a given confidence

        Args:
            utterance(str): conversational piece given by the user
            context(list): a list of entities
            N(int): number of results
        Returns: yield an object with the following fields
            utterance(str): the value passed in
            tags(list) : a list of tags found in utterance
            time(time) : duration since call of function
            confidence(float) : float indicating how confident of a match to the
                utterance. This might be used to determan the most likely intent.

        """
        context_trie = None
        if context and isinstance(context, list):
            # sort by confidence in ascending order, so
            # highest confidence for an entity is last.
            # see comment on TrieNode ctor
            context.sort(key=lambda x: x.get('confidence'))

            context_trie = Trie()
            for entity in context:
                entity_value, entity_type = entity.get('data')[0]
                context_trie.insert(entity_value.lower(),
                                    data=(entity_value, entity_type),
                                    weight=entity.get('confidence'))

        tagged = self._tagger.tag(utterance.lower(), context_trie=context_trie)
        start = time.time()
        bke = BronKerboschExpander(self._tokenizer)

        def score_clique(clique):
            score = 0.0
            for tagged_entity in clique:
                ec = tagged_entity.get('entities', [{'confidence': 0.0}])[0].get('confidence')
                score += ec * len(tagged_entity.get('entities', [{'match': ''}])[0].get('match')) / (
                    len(utterance) + 1)
            return score

        parse_results = bke.expand(tagged, clique_scoring_func=score_clique)
        count = 0
        for result in parse_results:
            count += 1
            parse_confidence = 0.0
            for tag in result:
                sample_entity = tag['entities'][0]
                entity_confidence = sample_entity.get('confidence', 0.0) * float(
                    len(sample_entity.get('match'))) / len(utterance)
                parse_confidence += entity_confidence
            yield {
                'utterance': utterance,
                'tags': result,
                'time': time.time() - start,
                'confidence': parse_confidence
            }

            if count >= N:
                break



# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import heapq
from ovos_adapt.entity_tagger import EntityTagger
from ovos_adapt.parser import Parser
from ovos_adapt.tools.text.tokenizer import EnglishTokenizer
from ovos_adapt.tools.text.trie import Trie

__author__ = 'seanfitz'


class IntentDeterminationEngine(object):
    """
    IntentDeterminationEngine

    The IntentDeterminationEngine is a greedy and naive implementation of intent determination. Given an utterance,
    it uses the Adapt parsing tools to come up with a sorted collection of tagged parses. A valid parse result contains
    no overlapping tagged entities, and its confidence is the sum of the tagged entity confidences, which are
    weighted based on the percentage of the utterance (per character) that the entity match represents.

    This system makes heavy use of generators to enable greedy algorithms to short circuit large portions of
    computation, however making use of context or regular expressions prevents these optimizations.
    """
    def __init__(self, tokenizer=None, trie=None):
        """
        Initialize the IntentDeterminationEngine

        Args:
            tokenizer(tokenizer) : tokenizer used to break up spoken text
                example EnglishTokenizer()
            trie(Trie): tree of matches to Entites
        """
        self.tokenizer = tokenizer or EnglishTokenizer()
        self.trie = trie or Trie()
        self.regular_expressions_entities = []
        self._regex_strings = set()
        self.intent_parsers = []

    def __best_intent(self, parse_result, context=[]):
        """
        For the specified parse_result, find the intent parser with the
        highest confidence match.

        Args:
            parse_result(list): results used to match the best intent.
            context(list): ?

        Returns:
            best_intent, best_tags:
                best_intent : The best intent for given results
                best_tags : The Tags for result
        """
        best_intent = None
        best_tags = None
        # TODO: there's a bunch of subtlety here around what the values of `match` and `key` should be
        # Longer term, this should probably be typed, barring any performance regressions.
        context_as_entities = [
            {
                'key': c['key'],
                'entities': [c],
                'from_context': True
             } for c in context
        ]
        for intent in self.intent_parsers:
            i, tags = intent.validate_with_tags(parse_result.get('tags') + context_as_entities, parse_result.get('confidence'))
            if not best_intent or (i and i.get('confidence') > best_intent.get('confidence')):
                best_intent = i
                best_tags = tags

        return best_intent, best_tags

    def __get_unused_context(self, parse_result, context):
        """ Used to get unused context from context.  Any keys not in
        parse_result

        Args:
            parse_results(list): parsed results used to identify what keys
                in the context are used.
            context(list): this is the context used to match with parsed results
                keys missing in the parsed results are the unused context

        Returns:
            list: A list of the unused context results.
        """
        tags_keys = set([t['key'] for t in parse_result['tags'] if t['from_context']])
        result_context = [c for c in context if c['key'] not in tags_keys]
        return result_context

    @property
    def tagger(self):
        return EntityTagger(self.trie, self.tokenizer,
                     self.regular_expressions_entities)

    def determine_intent(self, utterance, num_results=1, include_tags=False, context_manager=None):
        """
        Given an utterance, provide a valid intent.

        Args:
            utterance(str): an ascii or unicode string representing natural language speech
            include_tags(bool): includes the parsed tags (including position and confidence)
                as part of result
            context_manager(list): a context manager to provide context to the utterance
            num_results(int): a maximum number of results to be returned.

        Returns: A generator that yields dictionaries.
        """
        parser = Parser(self.tokenizer, self.tagger)

        context = []
        if context_manager:
            context = context_manager.get_context()

        # Adapt consumers assume that results are sorted by confidence. parser
        # will yield results sorted by utterance coverage, but regex
        # and context entities will have different weights, and
        # can influence final sorting.
        requires_final_sort = self.regular_expressions_entities or context

        def generate_intents():
            for result in parser.parse(utterance, N=num_results, context=context):
                # create a context without entities used in result
                remaining_context = self.__get_unused_context(result, context)
                best_intent, tags = self.__best_intent(result, remaining_context)
                if best_intent and best_intent.get('confidence', 0.0) > 0:
                    if include_tags:
                        best_intent['__tags__'] = tags
                    yield best_intent

        if requires_final_sort:
            sorted_iterable = sorted([
                i for i in generate_intents()
            ], key=lambda x: -x.get('confidence', 0.0))
        else:
            sorted_iterable = generate_intents()

        for intent in sorted_iterable:
            yield intent

    def register_entity(self, entity_value, entity_type, alias_of=None):
        """
        Register an entity to be tagged in potential parse results

        Args:
            entity_value(str): the value/proper name of an entity instance (Ex: "The Big Bang Theory")
            entity_type(str): the type/tag of an entity instance (Ex: "Television Show")
        """
        if alias_of:
            self.trie.insert(entity_value.lower(), data=(alias_of, entity_type))
        else:
            self.trie.insert(entity_value.lower(), data=(entity_value, entity_type))
            self.trie.insert(entity_type.lower(), data=(entity_type, 'Concept'))

    def register_regex_entity(self, regex_str):
        """
        A regular expression making use of python named group expressions.

        Example: (?P<Artist>.*)

        regex_str(str): a string representing a regular expression as defined above
        """
        if regex_str and regex_str not in self._regex_strings:
            self._regex_strings.add(regex_str)
            self.regular_expressions_entities.append(re.compile(regex_str, re.IGNORECASE))

    def register_intent_parser(self, intent_parser):
        """
        "Enforce" the intent parser interface at registration time.

        Args:
            intent_parser(intent): Intent to be registered.

        Raises:
            ValueError: on invalid intent
        """
        if hasattr(intent_parser, 'validate') and callable(intent_parser.validate):
            self.intent_parsers.append(intent_parser)
        else:
            raise ValueError("%s is not an intent parser" % str(intent_parser))

    def drop_intent_parser(self, parser_names):
        """Drop a registered intent parser.

        Arguments:
            parser_names (str or iterable): parser name to drop or list of
                                            names

        Returns:
            (bool) True if a parser was dropped else False
        """
        if isinstance(parser_names, str):
            parser_names = [parser_names]

        new_parsers = [p for p in self.intent_parsers
                       if p.name not in parser_names]
        num_original_parsers = len(self.intent_parsers)
        self.intent_parsers = new_parsers

        return len(self.intent_parsers) != num_original_parsers

    def drop_entity(self, entity_type=None, match_func=None):
        """Drop all entities mathching the given entity type or match function

        Arguments:
            entity_type (str): entity name to match against
            match_func (callable): match function to find entities

        Returns:
            (bool) True if vocab was found and removed otherwise False.
        """
        def default_match_func(data):
            return data and data[1] == entity_type

        ent_tuples = self.trie.scan(match_func or default_match_func)
        for entity in ent_tuples:
            self.trie.remove(*entity)

        return len(ent_tuples) != 0

    def drop_regex_entity(self, entity_type=None, match_func=None):
        """Remove registered regex entity.

        Arguments:
            entity_type (str): entity name to match against
            match_func (callable): match function to find entities

        Returns:
            (bool) True if vocab was found and removed otherwise False.
        """
        def default_match_func(regexp):
            return entity_type in regexp.groupindex.keys()

        match_func = match_func or default_match_func
        matches = [r for r in self.regular_expressions_entities
                   if match_func(r)]
        matching_patterns = [r.pattern for r in matches]

        matches = [
            r for r in self.regular_expressions_entities if r in matches
        ]
        for match in matches:
            self.regular_expressions_entities.remove(match)

        self._regex_strings = {
            r for r in self._regex_strings if r not in matching_patterns
        }

        return len(matches) != 0


class DomainIntentDeterminationEngine(object):
    """
    DomainIntentDeterminationEngine.

    The DomainIntentDeterminationEngine is a greedy and naive implementation of intent
    determination. Given an utterance, it uses the Adapt parsing tools to come up with a
    sorted collection of tagged parses. A valid parse result contains no overlapping
    tagged entities in a single domain, and its confidence is the sum of the tagged
    entity confidences, which are weighted based on the percentage of the utterance
    (per character) that the entity match represents.

    This system makes heavy use of generators to enable greedy algorithms to short circuit
    large portions of computation.
    """

    def __init__(self):
        """
        Initialize DomainIntentDeterminationEngine.
        """
        self.domains = {}

    @property
    def tokenizer(self):
        """
        A property to link into IntentEngine's tokenizer.

        Warning: this is only for backwards compatiblility and should not be used if you
            intend on using domains.

        Return: the domains tokenizer from its IntentEngine
        """
        domain = 0
        if domain not in self.domains:
            self.register_domain(domain=domain)
        return self.domains[domain].tokenizer

    @property
    def trie(self):
        """
        A property to link into IntentEngine's trie.

        warning:: this is only for backwards compatiblility and should not be used if you
        intend on using domains.

        Return: the domains trie from its IntentEngine
        """
        domain = 0
        if domain not in self.domains:
            self.register_domain(domain=domain)
        return self.domains[domain].trie

    @property
    def tagger(self):
        """
        A property to link into IntentEngine's intent_parsers.

        Warning: this is only for backwards compatiblility and should not be used if you
        intend on using domains.

        Return: the domains intent_parsers from its IntentEngine
        """
        domain = 0
        if domain not in self.domains:
            self.register_domain(domain=domain)
        return self.domains[domain].tagger

    @property
    def intent_parsers(self):
        """
        A property to link into IntentEngine's intent_parsers.

        Warning: this is only for backwards compatiblility and should not be used if you
            intend on using domains.

        Returns: the domains intent_parsers from its IntentEngine
        """
        domain = 0
        if domain not in self.domains:
            self.register_domain(domain=domain)
        return self.domains[domain].intent_parsers

    @property
    def _regex_strings(self):
        """
        A property to link into IntentEngine's _regex_strings.

        Warning: this is only for backwards compatiblility and should not be used if you
            intend on using domains.

        Returns: the domains _regex_strings from its IntentEngine
        """
        domain = 0
        if domain not in self.domains:
            self.register_domain(domain=domain)
        return self.domains[domain]._regex_strings

    @property
    def regular_expressions_entities(self):
        """
        A property to link into IntentEngine's regular_expressions_entities.

        Warning: this is only for backwards compatiblility and should not be used if you
            intend on using domains.

        Returns: the domains regular_expression_entities from its IntentEngine
        """
        domain = 0
        if domain not in self.domains:
            self.register_domain(domain=domain)
        return self.domains[domain].regular_expressions_entities

    def register_domain(self, domain=0, tokenizer=None, trie=None):
        """
        Register a domain with the intent engine.

        Args:
            tokenizer(tokenizer): The tokenizer you wish to use.
            trie(Trie): the Trie() you wish to use.
            domain(str): a string representing the domain you wish to add
        """
        self.domains[domain] = IntentDeterminationEngine(
            tokenizer=tokenizer, trie=trie)

    def register_entity(self, entity_value, entity_type, alias_of=None, domain=0):
        """
        Register an entity to be tagged in potential parse results.

        Args:
            entity_value(str): the value/proper name of an entity instance
                (Ex: "The Big Bang Theory")
            entity_type(str): the type/tag of an entity instance (Ex: "Television Show")
            domain(str): a string representing the domain you wish to add the entity to
        """
        if domain not in self.domains:
            self.register_domain(domain=domain)
        self.domains[domain].register_entity(entity_value=entity_value,
                                             entity_type=entity_type,
                                             alias_of=alias_of)

    def register_regex_entity(self, regex_str, domain=0):
        """
        A regular expression making use of python named group expressions.

        Example: (?P<Artist>.*)

        Args:
            regex_str(str): a string representing a regular expression as defined above
            domain(str): a string representing the domain you wish to add the entity to
        """
        if domain not in self.domains:
            self.register_domain(domain=domain)
        self.domains[domain].register_regex_entity(regex_str=regex_str)

    def determine_intent(self, utterance, num_results=1):
        """
        Given an utterance, provide a valid intent.

        utterance(str): an ascii or unicode string representing natural language speech
        num_results(int): a maximum number of results to be returned.

        Returns: A generator the yields dictionaries.
        """
        intents = []
        for domain in self.domains:
            gen = self.domains[domain].determine_intent(utterance=utterance,
                                                        num_results=1)
            for intent in gen:
                intents.append(intent)

        heapq.nlargest(
            num_results, intents, key=lambda domain: domain['confidence'])
        for intent in intents:
            yield intent

    def register_intent_parser(self, intent_parser, domain=0):
        """
        Register a intent parser with a domain.

        Args:
            intent_parser(intent): The intent parser you wish to register.
            domain(str): a string representing the domain you wish register the intent
                parser to.
        """
        if domain not in self.domains:
            self.register_domain(domain=domain)
        self.domains[domain].register_intent_parser(
            intent_parser=intent_parser)

    def drop_intent_parser(self, parser_names, domain):
        """Drop a registered intent parser.

        Arguments:
            parser_names (list, str): parser names to drop.
            domain (str): domain to drop from

        Returns:
            (bool) True if an intent parser was dropped else false.
        """
        return self.domains[domain].drop_intent_parser(parser_names)

    def drop_entity(self, domain, entity_type=None, match_func=None):
        """Drop all entities mathching the given entity type or match function.

        Arguments:
            domain (str): intent domain
            entity_type (str): entity name to match against
            match_func (callable): match function to find entities

        Returns:
            (bool) True if vocab was found and removed otherwise False.
        """
        return self.domains[domain].drop_entity(entity_type=entity_type,
                                                match_func=match_func)

    def drop_regex_entity(self, domain, entity_type=None, match_func=None):
        """Remove registered regex entity.

        Arguments:
            domain (str): intent domain
            entity_type (str): entity name to match against
            match_func (callable): match function to find entities

        Returns:
            (bool) True if vocab was found and removed otherwise False.
        """
        return self.domains[domain].drop_regex_entity(entity_type=entity_type,
                                                      match_func=match_func)



# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__author__ = 'seanfitz'


class TrieNode(object):
    def __init__(self, data=None, is_terminal=False):
        self.data = set()
        if data:
            self.data.add(data)
        self.is_terminal = is_terminal
        self.children = {}
        self.key = None
        # since weight is an attribute on the node (and not on the payload), all
        # payloads at this node have the weight of the last insert.
        self.weight = 1.0

    def lookup(self, iterable, index=0, gather=False, edit_distance=0, max_edit_distance=0, match_threshold=0.0, matched_length=0):
        """
        Args:
            iterable(hashable): a list of items used to traverse the Trie.
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]

            index(int): index of item for current position in traversal.
                we pass the original iterable and an index to avoid
                the cost of repeatedly copying the original iterable
            gather(bool): whether to return intermediate results (gather
                algorithm)
            edit_distance(int): current edit distance in the traversal.
            max_edit_distance(int): maximum edit distance
            match_threshold(float): minimum confidence of match for discovery
            matched_length(int): related to edit distance, for calculating
                confidence of match where
                confidence = (length - abs(matched_length - length)) / length

        yields:
            generator[TrieNode]: a generator that vends the results of the
                lookup, of  type TrieNode
        """
        if self.is_terminal:
            if index == len(iterable) or \
                    (gather and index < len(iterable) and iterable[index] == ' '):  # only gather on word break)
                confidence = float(len(self.key) - edit_distance) / float(max(len(self.key), index))
                if confidence > match_threshold:
                    yield {
                        'key': self.key,
                        'match': iterable[:index],
                        'data': self.data,
                        'confidence': confidence * self.weight
                    }

        if index < len(iterable) and iterable[index] in self.children:
            for result in self.children[iterable[index]]\
                    .lookup(iterable, index + 1, gather=gather,
                            edit_distance=edit_distance, max_edit_distance=max_edit_distance, matched_length=matched_length + 1):
                yield result

        # if there's edit distance remaining and it's possible to
        # match a word above the confidence threshold, continue searching
        potential_confidence = float(index - edit_distance + (max_edit_distance - edit_distance)) / \
                               (float(index) + (max_edit_distance - edit_distance)) if index + max_edit_distance - edit_distance > 0 else 0.0
        if edit_distance < max_edit_distance and potential_confidence > match_threshold:
            for child in list(self.children):
                if index >= len(iterable) or child != iterable[index]:
                    # substitution
                    for result in self.children[child]\
                        .lookup(iterable, index + 1, gather=gather,
                                edit_distance=edit_distance + 1, max_edit_distance=max_edit_distance, matched_length=matched_length):
                        yield result
                    # delete
                    for result in self.children[child]\
                        .lookup(iterable, index + 2, gather=gather,
                                edit_distance=edit_distance + 1, max_edit_distance=max_edit_distance, matched_length=matched_length):
                        yield result
                    # insert
                    for result in self.children[child]\
                        .lookup(iterable, index, gather=gather,
                                edit_distance=edit_distance + 1, max_edit_distance=max_edit_distance, matched_length=matched_length):
                        yield result

    def insert(self, iterable, index=0, data=None, weight=1.0):
        """Insert new node into tree

        Args:
            iterable(hashable): a list of items used to traverse the Trie.
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]
            data(object): data associated with the key
            index(int): an index used for insertion.
            weight(float): the wait given for the item added.
        """
        if index == len(iterable):
            self.is_terminal = True
            self.key = iterable
            self.weight = weight
            if data:
                self.data.add(data)
        else:
            if iterable[index] not in self.children:
                self.children[iterable[index]] = TrieNode()
            self.children[iterable[index]].insert(iterable, index + 1, data)

    def is_prefix(self, iterable, index=0):
        if index == len(iterable):
            return True
        elif iterable[index] in self.children:
            return self.children[iterable[index]].is_prefix(iterable, index + 1)
        else:
            return False

    def remove(self, iterable, data=None, index=0):
        """Remove an element from the trie

        Args
            iterable(hashable): a list of items used to traverse the Trie.
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]
            data(object): data associated with the key
            index(int): index of what is to me removed

        Returns:
        bool:
            True: if it was removed
            False: if it was not removed
        """
        if index == len(iterable):
            if self.is_terminal:
                if data:
                    self.data.remove(data)
                    if len(self.data) == 0:
                        self.is_terminal = False
                else:
                    self.data.clear()
                    self.is_terminal = False
                return True
            else:
                return False
        elif iterable[index] in self.children:
            return self.children[iterable[index]].remove(iterable, index=index+1, data=data)
        else:
            return False


class Trie(object):
    """Recursive implementation of a prefix trie (Trie)
    https://en.wikipedia.org/wiki/Trie
    Additionally supports #gather, a traversal whose results include
    any terminal nodes visited.

        Attributes:
            root(TrieNode): parent node to start the tree
            max_edit_distance(int): values > 0 allow for fuzzy matching
                with a maximum levenshtein edit distance
                https://en.wikipedia.org/wiki/Edit_distance
            match_threshold(float): only return values with a higher confidence
                than this value

    While most frequently used with strings, the Trie can be populated with any
    iterable (arrays of ints, arrays of objects, arrays of strings) as long
    as each value responds to `__hash__`.

    """

    def __init__(self, max_edit_distance=0, match_threshold=0.0):
        """Init the Trie object and create root node.

        Creates an Trie object with a root node with the passed in
        max_edit_distance and match_threshold.

        Args:
            max_edit_distance(int): values > 0 allow for fuzzy matching
                with a maximum levenshtein edit distance
                https://en.wikipedia.org/wiki/Edit_distance
            match_threshold(float): only return values with a higher confidence
                than this value
        """
        self.root = TrieNode('root')
        self.max_edit_distance = max_edit_distance
        self.match_threshold = match_threshold

    def gather(self, iterable):
        """Executes a "gather" traversal of the Trie
        Result set will include any `is_terminal` nodes encountered during
        the traversal

        Args:
            iterable(hashable): a list of items used to traverse the Trie
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]

        """
        for result in self.lookup(iterable, gather=True):
            yield result

    def lookup(self, iterable, gather=False):
        """Call the lookup on the root node with the given parameters.

        Args:
            iterable(hashable): a list of items used to traverse the Trie
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]
            gather(bool): flag to indicate whether gather results
                should be included

        """
        for result in self.root.lookup(iterable,
                                       gather=gather,
                                       edit_distance=0,
                                       max_edit_distance=self.max_edit_distance,
                                       match_threshold=self.match_threshold):
            yield result

    def insert(self, iterable, data=None, weight=1.0):
        """Used to insert into the trie

        Args:
            iterable(hashable): a list of items used to traverse the Trie
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]
            data(object): data to stored or merged for this iterable
        """
        self.root.insert(iterable, index=0, data=data, weight=weight)

    def remove(self, iterable, data=None):
        """Used to remove from the root node

        Args:
            iterable(hashable): a list of items used to traverse the Trie
                This represents the position of a node in the Trie, matching the
                iterable used at insertion time.
                For example:
                trie.insert('foo', {'bar': 'baz'})
                list(trie.lookup('foo')) == [TrieNode(data={'bar': 'baz'}, is_terminal=True)]
            data: data to removed. If None, or node is empty as a result,
                remove the node.
        """
        return self.root.remove(iterable, data=data)

    def scan(self, match_func):
        """Traverse the trie scanning for end nodes with matching data.

        Args:
            match_func (callable): function used to match data.

        Returns:
            (list) list with matching (data, value) pairs.
        """
        def _traverse(node, match_func, current=''):
            """Traverse Trie searching for nodes with matching data

            Performs recursive depth first search of Trie and collects
            value / data pairs matched by the match_func

            Arguments:
                node (trie node): Node to parse
                match_func (callable): Function performing match
                current (str): string "position" in Trie

            Returns:
                (list) list with matching (data, value) pairs.
            """
            # Check if node matches
            result = [(current, d) for d in node.data if match_func(d)]

            # Traverse further down into the tree
            for c in node.children:
                result += _traverse(node.children[c], match_func, current + c)
            return result

        return _traverse(self.root, match_func)



# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__author__ = 'seanfitz'

import itertools

CLIENT_ENTITY_NAME = 'Client'


def is_entity(tag, entity_name):
    for entity in tag.get('entities'):
        for v, t in entity.get('data'):
            if t.lower() == entity_name.lower():
                return True
    return False


def find_first_tag(tags, entity_type, after_index=-1):
    """Searches tags for entity type after given index

    Args:
        tags(list): a list of tags with entity types to be compared to
         entity_type
        entity_type(str): This is he entity type to be looking for in tags
        after_index(int): the start token must be greater than this.

    Returns:
        ( tag, v, confidence ):
            tag(str): is the tag that matched
            v(str): ? the word that matched?
            confidence(float): is a measure of accuracy.  1 is full confidence
                and 0 is none.
    """
    for tag in tags:
        for entity in tag.get('entities'):
            for v, t in entity.get('data'):
                if t.lower() == entity_type.lower() and \
                        (tag.get('start_token', 0) > after_index or \
                         tag.get('from_context', False)):
                    return tag, v, entity.get('confidence')

    return None, None, None


def find_next_tag(tags, end_index=0):
    for tag in tags:
        if tag.get('start_token') > end_index:
            return tag
    return None


def choose_1_from_each(lists):
    """
    The original implementation here was functionally equivalent to
    :func:`~itertools.product`, except that the former returns a generator
    of lists, and itertools returns a generator of tuples. This is going to do
    a light transform for now, until callers can be verified to work with
    tuples.

    Args:
        A list of lists or tuples, expected as input to
        :func:`~itertools.product`

    Returns:
        a generator of lists, see docs on :func:`~itertools.product`
    """
    for result in itertools.product(*lists):
        yield list(result)


def resolve_one_of(tags, at_least_one):
    """Search through all combinations of at_least_one rules to find a
    combination that is covered by tags

    Args:
        tags(list): List of tags with Entities to search for Entities
        at_least_one(list): List of Entities to find in tags

    Returns:
        object:
        returns None if no match is found but returns any match as an object
    """

    for possible_resolution in choose_1_from_each(at_least_one):
        resolution = {}
        pr = possible_resolution[:]
        for entity_type in pr:
            last_end_index = -1
            if entity_type in resolution:
                last_end_index = resolution[entity_type][-1].get('end_token')
            tag, value, c = find_first_tag(tags, entity_type,
                                           after_index=last_end_index)
            if not tag:
                break
            else:
                if entity_type not in resolution:
                    resolution[entity_type] = []
                resolution[entity_type].append(tag)
        # Check if this is a valid resolution (all one_of rules matched)
        if len(resolution) == len(possible_resolution):
            return resolution

    return None


class Intent(object):
    def __init__(self, name, requires, at_least_one, optional, excludes=None):
        """Create Intent object

        Args:
            name(str): Name for Intent
            requires(list): Entities that are required
            at_least_one(list): One of these Entities are required
            optional(list): Optional Entities used by the intent
        """
        self.name = name
        self.requires = requires
        self.at_least_one = at_least_one
        self.optional = optional
        self.excludes = excludes or []

    def validate(self, tags, confidence):
        """Using this method removes tags from the result of validate_with_tags

        Returns:
            intent(intent): Results from validate_with_tags
        """
        intent, tags = self.validate_with_tags(tags, confidence)
        return intent

    def validate_with_tags(self, tags, confidence):
        """Validate whether tags has required entites for this intent to fire

        Args:
            tags(list): Tags and Entities used for validation
            confidence(float): The weight associate to the parse result,
                as indicated by the parser. This is influenced by a parser
                that uses edit distance or context.

        Returns:
            intent, tags: Returns intent and tags used by the intent on
                failure to meat required entities then returns intent with
                confidence
                of 0.0 and an empty list for tags.
        """
        result = {'intent_type': self.name}
        intent_confidence = 0.0
        local_tags = tags[:]
        used_tags = []

        # Check excludes first
        for exclude_type in self.excludes:
            exclude_tag, _canonical_form, _tag_confidence = \
                find_first_tag(local_tags, exclude_type)
            if exclude_tag:
                result['confidence'] = 0.0
                return result, []

        for require_type, attribute_name in self.requires:
            required_tag, canonical_form, tag_confidence = \
                find_first_tag(local_tags, require_type)
            if not required_tag:
                result['confidence'] = 0.0
                return result, []

            result[attribute_name] = canonical_form
            if required_tag in local_tags:
                local_tags.remove(required_tag)
            used_tags.append(required_tag)
            intent_confidence += tag_confidence

        if len(self.at_least_one) > 0:
            best_resolution = resolve_one_of(local_tags, self.at_least_one)
            if not best_resolution:
                result['confidence'] = 0.0
                return result, []
            else:
                for key in best_resolution:
                    # TODO: at least one should support aliases
                    result[key] = best_resolution[key][0].get('key')
                    intent_confidence += \
                        1.0 * best_resolution[key][0]['entities'][0]\
                        .get('confidence', 1.0)
                used_tags.append(best_resolution[key][0])
                if best_resolution in local_tags:
                    local_tags.remove(best_resolution[key][0])

        for optional_type, attribute_name in self.optional:
            optional_tag, canonical_form, tag_confidence = \
                find_first_tag(local_tags, optional_type)
            if not optional_tag or attribute_name in result:
                continue
            result[attribute_name] = canonical_form
            if optional_tag in local_tags:
                local_tags.remove(optional_tag)
            used_tags.append(optional_tag)
            intent_confidence += tag_confidence

        total_confidence = (intent_confidence / len(tags) * confidence) \
            if tags else 0.0

        target_client, canonical_form, confidence = \
            find_first_tag(local_tags, CLIENT_ENTITY_NAME)

        result['target'] = target_client.get('key') if target_client else None
        result['confidence'] = total_confidence

        return result, used_tags


class IntentBuilder(object):
    """
    IntentBuilder, used to construct intent parsers.

    Attributes:
        at_least_one(list): A list of Entities where one is required.
            These are separated into lists so you can have one of (A or B) and
            then require one of (D or F).
        requires(list): A list of Required Entities
        optional(list): A list of optional Entities
        name(str): Name of intent

    Notes:
        This is designed to allow construction of intents in one line.

    Example:
        IntentBuilder("Intent")\
            .requires("A")\
            .one_of("C","D")\
            .optional("G").build()
    """
    def __init__(self, intent_name):
        """
        Constructor

        Args:
            intent_name(str): the name of the intents that this parser
            parses/validates
        """
        self.at_least_one = []
        self.requires = []
        self.excludes = []
        self.optional = []
        self.name = intent_name

    def one_of(self, *args):
        """
        The intent parser should require one of the provided entity types to
        validate this clause.

        Args:
            args(args): *args notation list of entity names

        Returns:
            self: to continue modifications.
        """
        self.at_least_one.append(args)
        return self

    def require(self, entity_type, attribute_name=None):
        """
        The intent parser should require an entity of the provided type.

        Args:
            entity_type(str): an entity type
            attribute_name(str): the name of the attribute on the parsed intent.
            Defaults to match entity_type.

        Returns:
            self: to continue modifications.
        """
        if not attribute_name:
            attribute_name = entity_type
        self.requires += [(entity_type, attribute_name)]
        return self

    def exclude(self, entity_type):
        """
        The intent parser must not contain an entity of the provided type.

        Args:
            entity_type(str): an entity type

        Returns:
            self: to continue modifications.
        """
        self.excludes.append(entity_type)
        return self

    def optionally(self, entity_type, attribute_name=None):
        """
        Parsed intents from this parser can optionally include an entity of the
         provided type.

        Args:
            entity_type(str): an entity type
            attribute_name(str): the name of the attribute on the parsed intent.
            Defaults to match entity_type.

        Returns:
            self: to continue modifications.
        """
        if not attribute_name:
            attribute_name = entity_type
        self.optional += [(entity_type, attribute_name)]
        return self

    def build(self):
        """
        Constructs an intent from the builder's specifications.

        :return: an Intent instance.
        """
        return Intent(self.name, self.requires,
                      self.at_least_one, self.optional,
                      self.excludes)
